"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = __importDefault(require("assert"));
const _1 = require(".");
const client = (0, _1.createClient)({
    cluster_id: 0n,
    replica_addresses: [process.env.TB_ADDRESS || '3000']
});
const Zeroed32Bytes = Buffer.alloc(32, 0);
const Zeroed48Bytes = Buffer.alloc(48, 0);
const accountA = {
    id: 17n,
    debits_pending: 0n,
    debits_posted: 0n,
    credits_pending: 0n,
    credits_posted: 0n,
    user_data_128: 0n,
    user_data_64: 0n,
    user_data_32: 0,
    reserved: 0,
    ledger: 1,
    code: 718,
    flags: 0,
    timestamp: 0n
};
const accountB = {
    id: 19n,
    debits_pending: 0n,
    debits_posted: 0n,
    credits_pending: 0n,
    credits_posted: 0n,
    user_data_128: 0n,
    user_data_64: 0n,
    user_data_32: 0,
    reserved: 0,
    ledger: 1,
    code: 719,
    flags: 0,
    timestamp: 0n
};
const tests = [];
function test(name, fn) {
    tests.push({ name, fn });
}
test.skip = (name, fn) => {
    console.log(name + ': SKIPPED');
};
test('id() monotonically increasing', async () => {
    let idA = (0, _1.id)();
    for (let i = 0; i < 10000000; i++) {
        if (i % 10000 == 0) {
            await new Promise(resolve => setTimeout(resolve, 1));
        }
        const idB = (0, _1.id)();
        assert_1.default.ok(idB > idA, 'id() returned an id that did not monotonically increase');
        idA = idB;
    }
});
test('range check `code` on Account to be u16', async () => {
    const account = { ...accountA, id: 0n };
    account.code = 65535 + 1;
    const codeError = await client.createAccounts([account]).catch(error => error);
    assert_1.default.strictEqual(codeError.message, 'code must be a u16.');
    const accounts = await client.lookupAccounts([account.id]);
    assert_1.default.deepStrictEqual(accounts, []);
});
test('can create accounts', async () => {
    const errors = await client.createAccounts([accountA]);
    assert_1.default.deepStrictEqual(errors, []);
});
test('can return error on account', async () => {
    const errors = await client.createAccounts([accountA, accountB]);
    assert_1.default.strictEqual(errors.length, 1);
    assert_1.default.deepStrictEqual(errors[0], { index: 0, result: _1.CreateAccountError.exists });
});
test('error if timestamp is not set to 0n on account', async () => {
    const account = { ...accountA, timestamp: 2n, id: 3n };
    const errors = await client.createAccounts([account]);
    assert_1.default.strictEqual(errors.length, 1);
    assert_1.default.deepStrictEqual(errors[0], { index: 0, result: _1.CreateAccountError.timestamp_must_be_zero });
});
test('can lookup accounts', async () => {
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    const account1 = accounts[0];
    assert_1.default.strictEqual(account1.id, 17n);
    assert_1.default.strictEqual(account1.credits_posted, 0n);
    assert_1.default.strictEqual(account1.credits_pending, 0n);
    assert_1.default.strictEqual(account1.debits_posted, 0n);
    assert_1.default.strictEqual(account1.debits_pending, 0n);
    assert_1.default.strictEqual(account1.user_data_128, 0n);
    assert_1.default.strictEqual(account1.user_data_64, 0n);
    assert_1.default.strictEqual(account1.user_data_32, 0);
    assert_1.default.strictEqual(account1.code, 718);
    assert_1.default.strictEqual(account1.ledger, 1);
    assert_1.default.strictEqual(account1.flags, 0);
    assert_1.default.ok(account1.timestamp > 0n);
    const account2 = accounts[1];
    assert_1.default.strictEqual(account2.id, 19n);
    assert_1.default.strictEqual(account2.credits_posted, 0n);
    assert_1.default.strictEqual(account2.credits_pending, 0n);
    assert_1.default.strictEqual(account2.debits_posted, 0n);
    assert_1.default.strictEqual(account2.debits_pending, 0n);
    assert_1.default.strictEqual(account2.user_data_128, 0n);
    assert_1.default.strictEqual(account2.user_data_64, 0n);
    assert_1.default.strictEqual(account2.user_data_32, 0);
    assert_1.default.strictEqual(account2.code, 719);
    assert_1.default.strictEqual(account2.ledger, 1);
    assert_1.default.strictEqual(account2.flags, 0);
    assert_1.default.ok(account2.timestamp > 0n);
});
test('can create a transfer', async () => {
    const transfer = {
        id: 1n,
        debit_account_id: accountB.id,
        credit_account_id: accountA.id,
        amount: 100n,
        user_data_128: 0n,
        user_data_64: 0n,
        user_data_32: 0,
        pending_id: 0n,
        timeout: 0,
        ledger: 1,
        code: 1,
        flags: 0,
        timestamp: 0n,
    };
    const errors = await client.createTransfers([transfer]);
    assert_1.default.deepStrictEqual(errors, []);
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    assert_1.default.strictEqual(accounts[0].credits_posted, 100n);
    assert_1.default.strictEqual(accounts[0].credits_pending, 0n);
    assert_1.default.strictEqual(accounts[0].debits_posted, 0n);
    assert_1.default.strictEqual(accounts[0].debits_pending, 0n);
    assert_1.default.strictEqual(accounts[1].credits_posted, 0n);
    assert_1.default.strictEqual(accounts[1].credits_pending, 0n);
    assert_1.default.strictEqual(accounts[1].debits_posted, 100n);
    assert_1.default.strictEqual(accounts[1].debits_pending, 0n);
});
test('can create a two-phase transfer', async () => {
    let flags = 0;
    flags |= _1.TransferFlags.pending;
    const transfer = {
        id: 2n,
        debit_account_id: accountB.id,
        credit_account_id: accountA.id,
        amount: 50n,
        user_data_128: 0n,
        user_data_64: 0n,
        user_data_32: 0,
        pending_id: 0n,
        timeout: 2e9,
        ledger: 1,
        code: 1,
        flags,
        timestamp: 0n,
    };
    const errors = await client.createTransfers([transfer]);
    assert_1.default.deepStrictEqual(errors, []);
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    assert_1.default.strictEqual(accounts[0].credits_posted, 100n);
    assert_1.default.strictEqual(accounts[0].credits_pending, 50n);
    assert_1.default.strictEqual(accounts[0].debits_posted, 0n);
    assert_1.default.strictEqual(accounts[0].debits_pending, 0n);
    assert_1.default.strictEqual(accounts[1].credits_posted, 0n);
    assert_1.default.strictEqual(accounts[1].credits_pending, 0n);
    assert_1.default.strictEqual(accounts[1].debits_posted, 100n);
    assert_1.default.strictEqual(accounts[1].debits_pending, 50n);
    const transfers = await client.lookupTransfers([transfer.id]);
    assert_1.default.strictEqual(transfers.length, 1);
    assert_1.default.strictEqual(transfers[0].id, 2n);
    assert_1.default.strictEqual(transfers[0].debit_account_id, accountB.id);
    assert_1.default.strictEqual(transfers[0].credit_account_id, accountA.id);
    assert_1.default.strictEqual(transfers[0].amount, 50n);
    assert_1.default.strictEqual(transfers[0].user_data_128, 0n);
    assert_1.default.strictEqual(transfers[0].user_data_64, 0n);
    assert_1.default.strictEqual(transfers[0].user_data_32, 0);
    assert_1.default.strictEqual(transfers[0].timeout > 0, true);
    assert_1.default.strictEqual(transfers[0].code, 1);
    assert_1.default.strictEqual(transfers[0].flags, 2);
    assert_1.default.strictEqual(transfers[0].timestamp > 0, true);
});
test('can post a two-phase transfer', async () => {
    let flags = 0;
    flags |= _1.TransferFlags.post_pending_transfer;
    const commit = {
        id: 3n,
        debit_account_id: BigInt(0),
        credit_account_id: BigInt(0),
        amount: 0n,
        user_data_128: 0n,
        user_data_64: 0n,
        user_data_32: 0,
        pending_id: 2n,
        timeout: 0,
        ledger: 1,
        code: 1,
        flags: flags,
        timestamp: 0n,
    };
    const errors = await client.createTransfers([commit]);
    assert_1.default.deepStrictEqual(errors, []);
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    assert_1.default.strictEqual(accounts[0].credits_posted, 150n);
    assert_1.default.strictEqual(accounts[0].credits_pending, 0n);
    assert_1.default.strictEqual(accounts[0].debits_posted, 0n);
    assert_1.default.strictEqual(accounts[0].debits_pending, 0n);
    assert_1.default.strictEqual(accounts[1].credits_posted, 0n);
    assert_1.default.strictEqual(accounts[1].credits_pending, 0n);
    assert_1.default.strictEqual(accounts[1].debits_posted, 150n);
    assert_1.default.strictEqual(accounts[1].debits_pending, 0n);
});
test('can reject a two-phase transfer', async () => {
    const transfer = {
        id: 4n,
        debit_account_id: accountB.id,
        credit_account_id: accountA.id,
        amount: 50n,
        user_data_128: 0n,
        user_data_64: 0n,
        user_data_32: 0,
        pending_id: 0n,
        timeout: 1e9,
        ledger: 1,
        code: 1,
        flags: _1.TransferFlags.pending,
        timestamp: 0n,
    };
    const transferErrors = await client.createTransfers([transfer]);
    assert_1.default.deepStrictEqual(transferErrors, []);
    const reject = {
        id: 5n,
        debit_account_id: BigInt(0),
        credit_account_id: BigInt(0),
        amount: 0n,
        user_data_128: 0n,
        user_data_64: 0n,
        user_data_32: 0,
        pending_id: 4n,
        timeout: 0,
        ledger: 1,
        code: 1,
        flags: _1.TransferFlags.void_pending_transfer,
        timestamp: 0n,
    };
    const errors = await client.createTransfers([reject]);
    assert_1.default.deepStrictEqual(errors, []);
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    assert_1.default.strictEqual(accounts[0].credits_posted, 150n);
    assert_1.default.strictEqual(accounts[0].credits_pending, 0n);
    assert_1.default.strictEqual(accounts[0].debits_posted, 0n);
    assert_1.default.strictEqual(accounts[0].debits_pending, 0n);
    assert_1.default.strictEqual(accounts[1].credits_posted, 0n);
    assert_1.default.strictEqual(accounts[1].credits_pending, 0n);
    assert_1.default.strictEqual(accounts[1].debits_posted, 150n);
    assert_1.default.strictEqual(accounts[1].debits_pending, 0n);
});
test('can link transfers', async () => {
    const transfer1 = {
        id: 6n,
        debit_account_id: accountB.id,
        credit_account_id: accountA.id,
        amount: 100n,
        user_data_128: 0n,
        user_data_64: 0n,
        user_data_32: 0,
        pending_id: 0n,
        timeout: 0,
        ledger: 1,
        code: 1,
        flags: _1.TransferFlags.linked,
        timestamp: 0n,
    };
    const transfer2 = {
        id: 6n,
        debit_account_id: accountB.id,
        credit_account_id: accountA.id,
        amount: 100n,
        user_data_128: 0n,
        user_data_64: 0n,
        user_data_32: 0,
        pending_id: 0n,
        timeout: 0,
        ledger: 1,
        code: 1,
        flags: 0,
        timestamp: 0n,
    };
    const errors = await client.createTransfers([transfer1, transfer2]);
    assert_1.default.strictEqual(errors.length, 2);
    assert_1.default.deepStrictEqual(errors[0], { index: 0, result: _1.CreateTransferError.linked_event_failed });
    assert_1.default.deepStrictEqual(errors[1], { index: 1, result: _1.CreateTransferError.exists_with_different_flags });
    const accounts = await client.lookupAccounts([accountA.id, accountB.id]);
    assert_1.default.strictEqual(accounts.length, 2);
    assert_1.default.strictEqual(accounts[0].credits_posted, 150n);
    assert_1.default.strictEqual(accounts[0].credits_pending, 0n);
    assert_1.default.strictEqual(accounts[0].debits_posted, 0n);
    assert_1.default.strictEqual(accounts[0].debits_pending, 0n);
    assert_1.default.strictEqual(accounts[1].credits_posted, 0n);
    assert_1.default.strictEqual(accounts[1].credits_pending, 0n);
    assert_1.default.strictEqual(accounts[1].debits_posted, 150n);
    assert_1.default.strictEqual(accounts[1].debits_pending, 0n);
});
test('cannot void an expired transfer', async () => {
    const transfer = {
        id: 6n,
        debit_account_id: accountB.id,
        credit_account_id: accountA.id,
        amount: 50n,
        user_data_128: 0n,
        user_data_64: 0n,
        user_data_32: 0,
        pending_id: 0n,
        timeout: 1,
        ledger: 1,
        code: 1,
        flags: _1.TransferFlags.pending,
        timestamp: 0n,
    };
    const transferErrors = await client.createTransfers([transfer]);
    assert_1.default.deepStrictEqual(transferErrors, []);
    await new Promise(_ => setTimeout(_, 1000));
    const reject = {
        id: 7n,
        debit_account_id: BigInt(0),
        credit_account_id: BigInt(0),
        amount: 0n,
        user_data_128: 0n,
        user_data_64: 0n,
        user_data_32: 0,
        pending_id: 6n,
        timeout: 0,
        ledger: 1,
        code: 1,
        flags: _1.TransferFlags.void_pending_transfer,
        timestamp: 0n,
    };
    const errors = await client.createTransfers([reject]);
    assert_1.default.strictEqual(errors.length, 1);
    assert_1.default.deepStrictEqual(errors[0], { index: 0, result: _1.CreateTransferError.pending_transfer_expired });
});
test('can get account transfers', async () => {
    const accountC = {
        id: 21n,
        debits_pending: 0n,
        debits_posted: 0n,
        credits_pending: 0n,
        credits_posted: 0n,
        user_data_128: 0n,
        user_data_64: 0n,
        user_data_32: 0,
        reserved: 0,
        ledger: 1,
        code: 718,
        flags: _1.AccountFlags.history,
        timestamp: 0n
    };
    const account_errors = await client.createAccounts([accountC]);
    assert_1.default.deepStrictEqual(account_errors, []);
    var transfers_created = [];
    for (var i = 0; i < 10; i++) {
        transfers_created.push({
            id: BigInt(i + 10000),
            debit_account_id: i % 2 == 0 ? accountC.id : accountA.id,
            credit_account_id: i % 2 == 0 ? accountB.id : accountC.id,
            amount: 100n,
            user_data_128: 0n,
            user_data_64: 0n,
            user_data_32: 0,
            pending_id: 0n,
            timeout: 0,
            ledger: 1,
            code: 1,
            flags: 0,
            timestamp: 0n,
        });
    }
    const transfers_created_result = await client.createTransfers(transfers_created);
    assert_1.default.deepStrictEqual(transfers_created_result, []);
    var filter = {
        account_id: accountC.id,
        timestamp_min: 0n,
        timestamp_max: 0n,
        limit: 8190,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits,
    };
    var transfers = await client.getAccountTransfers(filter);
    var account_balances = await client.getAccountBalances(filter);
    assert_1.default.strictEqual(transfers.length, transfers_created.length);
    assert_1.default.strictEqual(account_balances.length, transfers.length);
    var timestamp = 0n;
    var i = 0;
    for (var transfer of transfers) {
        assert_1.default.ok(timestamp < transfer.timestamp);
        timestamp = transfer.timestamp;
        assert_1.default.ok(account_balances[i].timestamp == transfer.timestamp);
        i++;
    }
    filter = {
        account_id: accountC.id,
        timestamp_min: 0n,
        timestamp_max: 0n,
        limit: 8190,
        flags: _1.AccountFilterFlags.debits | _1.AccountFilterFlags.reversed,
    };
    transfers = await client.getAccountTransfers(filter);
    account_balances = await client.getAccountBalances(filter);
    assert_1.default.strictEqual(transfers.length, transfers_created.length / 2);
    assert_1.default.strictEqual(account_balances.length, transfers.length);
    timestamp = 1n << 64n;
    i = 0;
    for (var transfer of transfers) {
        assert_1.default.ok(transfer.timestamp < timestamp);
        timestamp = transfer.timestamp;
        assert_1.default.ok(account_balances[i].timestamp == transfer.timestamp);
        i++;
    }
    filter = {
        account_id: accountC.id,
        timestamp_min: 0n,
        timestamp_max: 0n,
        limit: 8190,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.reversed,
    };
    transfers = await client.getAccountTransfers(filter);
    account_balances = await client.getAccountBalances(filter);
    assert_1.default.strictEqual(transfers.length, transfers_created.length / 2);
    assert_1.default.strictEqual(account_balances.length, transfers.length);
    timestamp = 1n << 64n;
    i = 0;
    for (var transfer of transfers) {
        assert_1.default.ok(transfer.timestamp < timestamp);
        timestamp = transfer.timestamp;
        assert_1.default.ok(account_balances[i].timestamp == transfer.timestamp);
        i++;
    }
    filter = {
        account_id: accountC.id,
        timestamp_min: 0n,
        timestamp_max: 0n,
        limit: transfers_created.length / 2,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits,
    };
    transfers = await client.getAccountTransfers(filter);
    account_balances = await client.getAccountBalances(filter);
    assert_1.default.strictEqual(transfers.length, transfers_created.length / 2);
    assert_1.default.strictEqual(account_balances.length, transfers.length);
    timestamp = 0n;
    i = 0;
    for (var transfer of transfers) {
        assert_1.default.ok(timestamp < transfer.timestamp);
        timestamp = transfer.timestamp;
        assert_1.default.ok(account_balances[i].timestamp == transfer.timestamp);
        i++;
    }
    filter = {
        account_id: accountC.id,
        timestamp_min: timestamp + 1n,
        timestamp_max: 0n,
        limit: transfers_created.length / 2,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits,
    };
    transfers = await client.getAccountTransfers(filter);
    account_balances = await client.getAccountBalances(filter);
    assert_1.default.strictEqual(transfers.length, transfers_created.length / 2);
    assert_1.default.strictEqual(account_balances.length, transfers.length);
    i = 0;
    for (var transfer of transfers) {
        assert_1.default.ok(timestamp < transfer.timestamp);
        timestamp = transfer.timestamp;
        assert_1.default.ok(account_balances[i].timestamp == transfer.timestamp);
        i++;
    }
    filter = {
        account_id: accountC.id,
        timestamp_min: timestamp + 1n,
        timestamp_max: 0n,
        limit: transfers_created.length / 2,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits,
    };
    transfers = await client.getAccountTransfers(filter);
    account_balances = await client.getAccountBalances(filter);
    assert_1.default.deepStrictEqual(transfers, []);
    assert_1.default.strictEqual(account_balances.length, transfers.length);
    filter = {
        account_id: accountC.id,
        timestamp_min: 0n,
        timestamp_max: 0n,
        limit: transfers_created.length / 2,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits | _1.AccountFilterFlags.reversed,
    };
    transfers = await client.getAccountTransfers(filter);
    account_balances = await client.getAccountBalances(filter);
    assert_1.default.strictEqual(transfers.length, transfers_created.length / 2);
    assert_1.default.strictEqual(account_balances.length, transfers.length);
    timestamp = 1n << 64n;
    i = 0;
    for (var transfer of transfers) {
        assert_1.default.ok(timestamp > transfer.timestamp);
        timestamp = transfer.timestamp;
        assert_1.default.ok(account_balances[i].timestamp == transfer.timestamp);
        i++;
    }
    filter = {
        account_id: accountC.id,
        timestamp_min: 0n,
        timestamp_max: timestamp - 1n,
        limit: transfers_created.length / 2,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits | _1.AccountFilterFlags.reversed,
    };
    transfers = await client.getAccountTransfers(filter);
    account_balances = await client.getAccountBalances(filter);
    assert_1.default.strictEqual(transfers.length, transfers_created.length / 2);
    assert_1.default.strictEqual(account_balances.length, transfers.length);
    i = 0;
    for (var transfer of transfers) {
        assert_1.default.ok(timestamp > transfer.timestamp);
        timestamp = transfer.timestamp;
        assert_1.default.ok(account_balances[i].timestamp == transfer.timestamp);
        i++;
    }
    filter = {
        account_id: accountC.id,
        timestamp_min: 0n,
        timestamp_max: timestamp - 1n,
        limit: transfers_created.length / 2,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits | _1.AccountFilterFlags.reversed,
    };
    transfers = await client.getAccountTransfers(filter);
    account_balances = await client.getAccountBalances(filter);
    assert_1.default.deepStrictEqual(transfers, []);
    assert_1.default.strictEqual(account_balances.length, transfers.length);
    filter = {
        account_id: 0n,
        timestamp_min: 0n,
        timestamp_max: 0n,
        limit: 8190,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits,
    };
    assert_1.default.deepStrictEqual((await client.getAccountTransfers(filter)), []);
    assert_1.default.deepStrictEqual((await client.getAccountBalances(filter)), []);
    filter = {
        account_id: accountC.id,
        timestamp_min: (1n << 64n) - 1n,
        timestamp_max: 0n,
        limit: 8190,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits,
    };
    assert_1.default.deepStrictEqual((await client.getAccountTransfers(filter)), []);
    assert_1.default.deepStrictEqual((await client.getAccountBalances(filter)), []);
    filter = {
        account_id: accountC.id,
        timestamp_min: 0n,
        timestamp_max: (1n << 64n) - 1n,
        limit: 8190,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits,
    };
    assert_1.default.deepStrictEqual((await client.getAccountTransfers(filter)), []);
    assert_1.default.deepStrictEqual((await client.getAccountBalances(filter)), []);
    filter = {
        account_id: accountC.id,
        timestamp_min: (1n << 64n) - 2n,
        timestamp_max: 1n,
        limit: 8190,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits,
    };
    assert_1.default.deepStrictEqual((await client.getAccountTransfers(filter)), []);
    assert_1.default.deepStrictEqual((await client.getAccountBalances(filter)), []);
    filter = {
        account_id: accountC.id,
        timestamp_min: 0n,
        timestamp_max: 0n,
        limit: 0,
        flags: _1.AccountFilterFlags.credits | _1.AccountFilterFlags.debits,
    };
    assert_1.default.deepStrictEqual((await client.getAccountTransfers(filter)), []);
    assert_1.default.deepStrictEqual((await client.getAccountBalances(filter)), []);
    filter = {
        account_id: accountC.id,
        timestamp_min: 0n,
        timestamp_max: 0n,
        limit: 8190,
        flags: _1.AccountFilterFlags.none,
    };
    assert_1.default.deepStrictEqual((await client.getAccountTransfers(filter)), []);
    assert_1.default.deepStrictEqual((await client.getAccountBalances(filter)), []);
    filter = {
        account_id: accountC.id,
        timestamp_min: 0n,
        timestamp_max: 0n,
        limit: 8190,
        flags: 0xFFFF,
    };
    assert_1.default.deepStrictEqual((await client.getAccountTransfers(filter)), []);
    assert_1.default.deepStrictEqual((await client.getAccountBalances(filter)), []);
});
async function main() {
    const start = new Date().getTime();
    try {
        for (let i = 0; i < tests.length; i++) {
            await tests[i].fn().then(() => {
                console.log(tests[i].name + ": PASSED");
            }).catch(error => {
                console.log(tests[i].name + ": FAILED");
                throw error;
            });
        }
        const end = new Date().getTime();
        console.log('Time taken (s):', (end - start) / 1000);
    }
    finally {
        await client.destroy();
    }
}
main().catch((error) => {
    console.log('operator:', error.operator);
    console.log('stack:', error.stack);
    process.exit(-1);
});
//# sourceMappingURL=test.js.map